"""
Definition of the Result type and accompanying classes
"""

__all__ = [
    "Result",
    "Ok",
    "Err",
    "UnwrapError",
    ]

import inspect

from typing import TypeVar, Union, Tuple, Literal, Optional, Callable
from typing_extensions import TypeGuard

T = TypeVar("T")
U = TypeVar("U")
E = TypeVar("E")

OkT = Tuple[Literal["ok"], T]
ErrT = Tuple[Literal["err"], E]

Result = Union[OkT[T], ErrT[E]]


class UnwrapError(Exception):
    """
    Error that is raised when unwrapping a Result fails
    """
    __slots__ = ()


# ClassOk and ClassErr both contain overloaded methods for get and map,
#  these methods are currently commented out due to breaking mypy
# issue: https://github.com/python/mypy/issues/7781
#
# when issue is marked patched, the overloads should be uncommented
#
# this pylint pragma is set to ignore the string comments,
#  it should also be removed once patched
# pylint: disable=pointless-string-statement


class ClassOk:
    """
    Constructor and inspector class for getting the Ok variant of a Result type
    """
    __slots__ = ()

    @staticmethod
    def __call__(value: T, /) -> OkT[T]:
        """
        Wraps a value in an Ok type
        """
        return ("ok", value)

    @staticmethod
    def is_instance(result: Result[T, E], /) -> TypeGuard[OkT[T]]:
        """
        Returns True if Result is an Ok variant
        """
        return result[0] == "ok"

    # See comment above, uncomment when patched
    """
    @overload
    @staticmethod
    def get(result: OkT[T], /) -> T:
        ...

    @overload
    @staticmethod
    def get(result: ErrT[E], /) -> None:
        ...

    @overload
    @staticmethod
    def get(result: Result[T, E], /) -> Optional[T]:
        ...
    """

    @staticmethod
    def get(result: Result[T, E], /) -> Optional[T]:
        """
        Returns Ok variant from Result or returns None
        """
        if result[0] == "err":
            return None

        return result[1]

    @staticmethod
    def unwrap(result: Result[T, E], /) -> T:
        """
        Returns Ok variant from Result or raises
        """
        if result[0] == "err":
            # print lineno and fname of caller
            if (cur_frame := inspect.currentframe()
                ) is not None and (caller_frame := cur_frame.f_back) is not None:
                loc = f"{caller_frame.f_code.co_filename}:{caller_frame.f_lineno} "
            else:
                loc = ""

            raise UnwrapError(
                f"{loc}Result type was Err variant, expected Ok variant\n"
                f"Err: {type(result[1]).__name__}='{result[1]}'"
                )

        return result[1]

    # See comment above, uncomment when patched
    """
    @overload
    @staticmethod
    def map(result: OkT[T], call: Callable[[T], U]) -> OkT[U]:
        ...

    @overload
    @staticmethod
    def map(result: ErrT[E], call: Callable[[T], U]) -> ErrT[E]:
        ...

    @overload
    @staticmethod
    def map(result: Result[T, E], call: Callable[[T], U]) -> Result[U, E]:
        ...
    """

    @staticmethod
    def map(result: Result[T, E], call: Callable[[T], U]) -> Result[U, E]:
        """
        Maps Ok value of result with function, or returns Err variant unchanged
        """
        if Ok.is_instance(result):
            return Ok(call(result[1]))
        if Err.is_instance(result):
            return result

        raise RuntimeError("Unreachable!")


class ClassErr:
    """
    Constructor and inspector class for getting the Err variant of a Result type
    """
    __slots__ = ()

    @staticmethod
    def __call__(value: E, /) -> ErrT[E]:
        """
        Wraps a value in an Err type
        """
        return ("err", value)

    @staticmethod
    def is_instance(result: Result[T, E], /) -> TypeGuard[ErrT[E]]:
        """
        Returns True if Result is an Err variant
        """
        return result[0] == "err"

    # See comment above, uncomment when patched
    """
    @staticmethod
    @overload
    def get(result: ErrT[E], /) -> E:
        ...

    @staticmethod
    @overload
    def get(result: OkT[T], /) -> None:
        ...

    @staticmethod
    @overload
    def get(result: Result[T, E], /) -> Optional[E]:
        ...
    """

    @staticmethod
    def get(result: Result[T, E], /) -> Optional[E]:
        """
        Returns Err variant from Result or returns None
        """
        if result[0] == "ok":
            return None

        return result[1]

    @staticmethod
    def unwrap(result: Result[T, E], /) -> E:
        """
        Returns Err variant of Result type or raises
        """
        if result[0] == "ok":
            # print lineno and fname of caller
            if (cur_frame := inspect.currentframe()
                ) is not None and (caller_frame := cur_frame.f_back) is not None:
                loc = f"{caller_frame.f_code.co_filename}:{caller_frame.f_lineno} "
            else:
                loc = ""

            raise UnwrapError(
                f"{loc}Result type was Ok variant, expected Err variant\n"
                f"Ok: {type(result[1]).__name__}='{result[1]}'"
                )

        return result[1]

    # See comment above, uncomment when patched
    """
    @overload
    @staticmethod
    def map(result: ErrT[E], call: Callable[[E], U]) -> ErrT[U]:
        ...

    @overload
    @staticmethod
    def map(result: OkT[T], call: Callable[[E], U]) -> OkT[T]:
        ...

    @overload
    @staticmethod
    def map(result: Result[T, E], call: Callable[[E], U]) -> Result[T, U]:
        ...
    """

    @staticmethod
    def map(result: Result[T, E], call: Callable[[E], U]) -> Result[T, U]:
        """
        Maps Err value of result with function, or returns Ok variant unchanged
        """
        if Err.is_instance(result):
            return Err(call(result[1]))
        if Ok.is_instance(result):
            return result

        raise RuntimeError("Unreachable!")


Ok: ClassOk = ClassOk()
Err: ClassErr = ClassErr()



"""
Strips a url
"""
import copy

from ._special_cases import special_cases_map
from ._result import Result, Err, Ok
from ._types import HttpUrl, UrlError
from .testing import test


def strip_last_query(url: HttpUrl, /) -> Result[HttpUrl, UrlError]:
    """
    Strips all but the first query of a url, works for general cases
    """

    url = copy.copy(url)

    url.query = url.query[:1]

    return Ok(url)


def strip_url(url_str: str, /) -> Result[HttpUrl, UrlError]:
    """
    Strips a url of tracking elements and bloat
    """

    if not url_str.startswith("http"):
        return Err(UrlError("Url does not start with http(s)"))

    parsed: Result[HttpUrl, UrlError] = HttpUrl.from_str(url_str)

    if (err := Err.get(parsed)) is not None:
        return Err(err)

    url = Ok.unwrap(parsed)

    if url.domain in special_cases_map:
        return special_cases_map[url.domain](url)

    return strip_last_query(url)


@test
def test_error_url() -> None:
    """
    Tests url parsing erroring correctly
    """

    assert Err.is_instance(strip_url("s\\gfadhbgjdkshfgb"))



# main file
"""
A small library to strip urls of tracking elements, with easy extensibility
"""

__all__ = [
    "Result",
    "Ok",
    "Err",
    "HttpUrl",
    "UrlError",
    "UrlParseError",
    "StripFunc",
    "StripFuncResult",
    "register",
    "strip_url",
    "special_cases",
    ]

import typing as _typing

from ._result import Result, Ok, Err
from ._types import HttpUrl, UrlError, UrlParseError, StripFunc, StripFuncResult
from ._special_cases import register, special_cases
from ._strip import strip_url


class _VersionInfo(_typing.NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: _typing.Literal["alpha", "beta", "candidate", "final"]
    serial: int

    def as_pep440_str(self) -> str:
        """
        Formats a _VersionInfo into a pep440 compliant a-b-rc string
        """
        _table = {"alpha": "a", "beta": "b", "candidate": "rc", "final": ""}
        serial = str(self.serial) if self.serial else ""
        return f"{self.major}.{self.minor}.{self.micro}{_table[self.releaselevel]}{serial}"


version_info = _VersionInfo(0, 2, 0, "final", 0)
__version__ = version_info.as_pep440_str()



"""
Testing library for url_strip that allows easily writing and running tests
"""
import typing as _typing
import traceback as _traceback

if _typing.TYPE_CHECKING:
    Func = _typing.TypeVar("Func", bound=_typing.Callable[[], None])

__all__ = [
    "test",
    "run_tests",
    ]

_test_list: _typing.List[_typing.Callable[[], None]] = []


def test(c: "Func", /) -> "Func":
    """
    Decorator to add a function to the testing library
    """

    _test_list.append(c)

    return c


def run_tests() -> int:
    """
    Runs all tests in the list, printing any exceptions, with the return int
     indicating how many tests failed.
    0 may be treated as no tests failing, or success.
    """

    failure = 0

    for i in _test_list:
        try:
            i()
        except Exception as err:  # pylint: disable=broad-except
            # allow catching broad exceptions here to allow running all tests
            _traceback.print_exception(type(err), err, err.__traceback__)
            failure += 1

    return failure



"""
Special cases init handler and predefined domains
"""
from typing import Dict, TypeVar, Callable, Iterator, List, Union

from ._result import Ok, Err, Result
from ._types import StripFunc, StripFuncResult, HttpUrl, UrlError
from .testing import test

special_cases_map: Dict[str, StripFunc] = {}

F = TypeVar("F", bound=StripFunc)


class NoDomainError(Exception):
    """
    Exception raised when a function is passed to register with no domains
    """
    __slots__ = ()


def special_cases() -> Iterator[str]:
    """
    Returns an iterator of all special case domain names
    """
    return iter(special_cases_map)


def register(*, domain: Union[str, List[str]]) -> Callable[[F], F]:
    """
    Register a stripping function to run for a specific domain
    """
    def deco(f: F, /) -> F:
        if isinstance(domain, str):
            special_cases_map[domain] = f
        else:
            if len(domain) == 0:
                raise NoDomainError(f"Function {f} registered with no domains")
            for i in domain:
                special_cases_map[i] = f
        return f

    return deco


def _no_query(v: HttpUrl, /) -> HttpUrl:
    return HttpUrl(v.domain, v.path, [], v.fragment)


def _takes_str(c: Callable[[HttpUrl], StripFuncResult],
               /) -> Callable[[str], Result[str, UrlError]]:
    def takes_str(s: str, /) -> Result[str, UrlError]:
        return Ok.map(c(Ok.unwrap(HttpUrl.from_str(s))), HttpUrl.into_str)

    return takes_str


@register(domain=["youtube.com", "www.youtube.com"])
def youtube_strip(v: HttpUrl, /) -> StripFuncResult:
    """
    Strip function for youtube domains
    """

    quer = dict(v.query)

    if v.path == "/watch":

        if "v" not in quer:
            return Err(UrlError("Address was a watch link, but could not find video id"))

        return Ok(HttpUrl("youtu.be", f"/{quer['v']}", [], v.fragment))

    return Ok(_no_query(v))


@test
def test_yt_url() -> None:
    """
    Tests youtube stripfunc
    """
    stripfunc = _takes_str(youtube_strip)
    assert stripfunc("https://youtube.com/watch?v=abcdefg1234&tracker=345345"
                     ) == Ok("https://youtu.be/abcdefg1234")


@register(domain=["www.amazon.com", "www.amazon.co.uk"])
def amazon_strip(v: HttpUrl, /) -> StripFuncResult:
    """
    Strip function for amazon domains
    """

    splitpath = v.path.split("/")

    if (idx := splitpath.index("dp")) != -1:
        try:
            product_id = splitpath[idx + 1]
        except IndexError:
            return Ok(_no_query(v))

        return Ok(HttpUrl(v.domain, f"/dp/{product_id}", [], v.fragment))

    return Ok(_no_query(v))


@test
def test_amazon_url() -> None:
    """
    Tests amazon urls
    """
    stripfunc = _takes_str(amazon_strip)

    # real amazon url btw
    res = stripfunc(
        "https://www.amazon.com/KONNWEI-Bluetooth-Wireless-Diagnostic-compatible/dp/B073F57QT3/"
        "?content-id=amzn1.sym.bbb6bbd8-d236-47cb-b42f-734cb0cacc1f"
        )

    assert res == Ok("https://www.amazon.com/dp/B073F57QT3")


@register(domain="twitter.com")
@register(domain="www.reddit.com")
@register(domain=["ebay.com", "www.ebay.com", "www.ebay.co.uk"])
@register(domain=["www.tiktok.com", "vm.tiktok.com"])
def no_queryable(v: HttpUrl, /) -> StripFuncResult:
    """
    Strips collection of domains that can be effectively stripped by
     removing the query string on them
    """

    return Ok(_no_query(v))


@test
def test_no_query() -> None:
    """
    Tests that no_queryable removes queries
    """

    func = _takes_str(no_queryable)

    assert func("https://google.com/search?v=among") == Ok("https://google.com/search")

    assert Ok.is_instance(
        func("https://www.tiktok.com/@amqsinc/video/7118107323461094699?_t=8TshIG8gDyQ&_r=1")
        )



from typing import List
from setuptools import setup, find_packages


def get_requires() -> List[str]:
    return ["typing_extensions >= 3.10.0.0, < 5"]


def support_pyvers(major: int, minor: range) -> List[str]:
    return [f"Programming Language :: Python :: {major}.{i}" for i in minor]


setup(
    name="url_strip",
    version="0.2.0",
    description=
    "Library to strip http urls of tracking elements, and use shorthand variants of urls",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    author="ultrabear",
    author_email="bearodark@gmail.com",
    packages=find_packages(),
    install_requires=get_requires(),
    extras_require={
        # yapf requires toml to load pyproject.toml
        "dev": ["mypy", "pylint", "yapf", "toml", "pyflakes", "pyright"],
        },
    python_requires=">=3.8.0",
    classifiers=[
        *support_pyvers(3, range(8, 11)), "Programming Language :: Python :: 3 :: Only",
        "Development Status :: 4 - Beta", "Intended Audience :: Developers",
        "Natural Language :: English", "Operating System :: OS Independent",
        "Topic :: Text Processing :: Filters", "Typing :: Typed",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
        ],
    )



"""
Base types for the url_strip library
"""

from dataclasses import dataclass
import io
from typing import Callable, Optional, Tuple, List, Type, Final

import re

from ._result import Result, Ok, Err

# chars unreserved (allowed) in a url and the expansion char
# based on https://www.ietf.org/rfc/rfc3986.txt
# added @ to allow capturing tiktok urls
URL_CHARS: Final = r"([a-zA-z0-9\-\.\_\-@]|%[0-9a-fA-F]{2})"

DOMAIN_CAPTURE: Final = f"(?P<domain>{URL_CHARS}+)"

# Colons are allowed in the path
PATH_CAPTURE: Final = f"(?P<path>(/({URL_CHARS}|\\:)*)+)"

QUERY_KV: Final = f"{URL_CHARS}+={URL_CHARS}+"
QUERY_CAPTURE: Final = f"(\\?(?P<query>{QUERY_KV}(&{QUERY_KV})*))"

FRAGMENT_CAPTURE: Final = f"(?P<fragment>#{URL_CHARS}+)"

# This regex will capture something that is intended to be a url
URL_REGEX: Final = re.compile(
    f"\\b(?:https?://){DOMAIN_CAPTURE}{PATH_CAPTURE}?{QUERY_CAPTURE}?{FRAGMENT_CAPTURE}?\\b"
    )


class UrlError(Exception):
    """
    A url related error
    """
    __slots__ = ()


class UrlParseError(UrlError):
    """
    An error relating to a url parse
    """
    __slots__ = ()


@dataclass
class HttpUrl:
    """
    An HttpUrl is a dataclass of a url representing its base parts, making it easier to modify
    """
    __slots__ = "domain", "path", "query", "fragment"
    domain: str
    path: str
    query: List[Tuple[str, str]]
    fragment: Optional[str]

    @staticmethod
    def _parse_query_str(query_str: Optional[str], /) -> List[Tuple[str, str]]:
        """
        Parses a query string into its ast value
        """
        query: List[Tuple[str, str]] = []

        if query_str is not None:
            for i in query_str.split("&"):
                key, value = i.split("=")
                query.append((key, value), )

        return query

    @staticmethod
    def _expand_format_chars(s: str, /) -> Result[str, UrlParseError]:
        """
        Expands url format chars in a string
        """

        output = io.BytesIO()
        skip = 0

        for idx, char in enumerate(s):
            if skip != 0:
                skip -= 1
            elif char == "%":
                skip = 2
                val = s[idx + 1:idx + 3]
                if len(val) != 2:
                    return Err(
                        UrlParseError(
                            "Could not expand url element, not enough chars after % symbol"
                            )
                        )

                try:
                    data = int(val, 16)
                except ValueError:
                    print(val)
                    return Err(
                        UrlParseError(
                            "Could not expand url element, chars after % symbol are not hex encoded"
                            )
                        )

                output.write(bytes([data]))

            else:
                output.write(char.encode("utf8"))

        try:
            return Ok(output.getvalue().decode("utf8"))
        except UnicodeDecodeError as err:
            return Err(UrlParseError(err))

    def into_str(self, *, protocol: str = "https") -> str:
        """
        Returns a string representation of the HttpUrl
        """
        query_str = "?" + "&".join(f"{k}={v}" for k, v in self.query) if self.query else ""
        fragment_str = self.fragment if self.fragment is not None else ""

        return f"{protocol}://{self.domain}{self.path}{query_str}{fragment_str}"

    def path_decoded(self) -> Result[List[str], UrlError]:
        """
        Expands % format characters and returns path as a list
        """
        splitpath = self.path.split("/")

        out: List[str] = []

        for i in map(self._expand_format_chars, splitpath):

            if (val := Ok.get(i)) is not None:
                out.append(val)
            else:
                return Err(Err.unwrap(i))

        return Ok(out)

    @classmethod
    def from_str(cls: Type["HttpUrl"], url: str, /) -> Result["HttpUrl", UrlParseError]:
        """
        Parses an HttpUrl from a string
        """

        # remove any trailing path pieces
        url = url.rstrip("/")

        if (capture := URL_REGEX.match(url)) is None:
            return Err(UrlParseError("Regex could not find url match"))

        # Require entire field to be url
        if capture.span() != (0, len(url)):
            return Err(UrlParseError("Regex capture does not span entire string passed"))

        if (domain := capture.group("domain")) is None:
            return Err(UrlParseError("Could not capture domain group"))

        if (path := capture.group("path")) is None:
            path = ""

        query = cls._parse_query_str(capture.group("query"))

        fragment = capture.group("fragment")

        return Ok(cls(domain, path, query, fragment))


StripFuncResult = Result[HttpUrl, UrlError]
StripFunc = Callable[[HttpUrl], StripFuncResult]
